!
!   FE mesh I/O   (FE formats and conversion)  
!   
!
!   Last edited by BP on 10/6/03  
!
!    vol (netgen format)
!    neu (neutral format)
!    dfp (diffpack format)
!    tec (tecplot format)
!    
! *** Methods: 
!
!   VOL_read_FE (unit, F2C, F2N) 
!   NEU_read_FE (unit, F2C, F2N)
!*   DFP_read_FE (unit, F2C, F2N)
!
!   TEC_write_FE (unit, NArray, F2C, F2N)
!   
!   get_F2C(C2N, F2C)
!   get_F2N(C2N, F2N)
!
!   get_C2N(F2C,F2N)
!
!
Module FE_mesh_IO
Implicit none

! Inputs
Integer :: ND, NN
Real, Allocatable, Dimension(:,:) :: npos

Integer :: NCN, NC
Integer, Allocatable, Dimension(:) :: BC_Cell
Integer, Allocatable, Dimension(:,:) :: C2N

! Face Connectivity
Integer :: NFN, NF
Integer, Allocatable, Dimension(:,:) :: F2C, F2N

! BCs
Integer :: NFBC, NNBC
Integer, Allocatable, Dimension(:,:) :: nBC_pt, fBC_pt
Integer, Allocatable, Dimension(:)   :: BC_Node, BC_Face
  		
contains 

!
!===============================================================
!===============================================================
!  get:   ND,NN,npos, NC,C2N,BC_Cell
!
Subroutine VOL_read_FE (filename)
Implicit None
Character (Len=*) :: filename

Integer :: ios,unit, i,j
Integer :: i1,i2,i3,i4,j1,j2,j3, NE,NBF

    unit = 10
    Open (unit, File=filename//".vol", Iostat=ios, Form="Formatted", Action="Read")
	if (ios /= 0) then
		Print *, "Error opening .vol file. Error code = ", ios, filename//".vol" 
		stop 
	end if


	Read(unit,*)			! name
	Read(unit,*)
	Read(unit,*) ND

	Read(unit,*)			! B faces (cells in 2D)		
	Read(unit,*)
	Read(unit,*)
	if (ND == 2) then
	  Read(unit,*) NC
	  NCN = ND+1
	  Allocate ( C2N(NCN,NC), BC_Cell(NC) )
      do i=1,NC
	     Read(unit,*) i1, BC_Cell(i), i2, i3, i4, (C2N(j,i), j=1,NCN), j1, j2, j3
	  end do
	else	
	  Read(unit,*) NBF
      do i=1,NBF
	     Read(unit,*) 
	  end do
    end if

	Read(unit,*)			! cell elements			
	Read(unit,*)
	Read(unit,*)
	Read(unit,*) 
	if (ND == 3) then
	  Read(unit,*) NC
      NCN = ND + 1
	  Allocate( C2N(NCN,NC), BC_Cell(NC) )
      do i=1,NC
	     Read(unit,*) BC_Cell(i), i1, (C2N(j,i), j=1,NCN) 
	  end do
    else
	  Read(unit,*) i1
	  if (i1 /= 0) Write(*,*)"Trouble "	 
    end if

	Read(unit,*)         ! B edges		
	Read(unit,*)
	Read(unit,*)
	Read(unit,*)
	Read(unit,*) NE
    do i=1,NE
	   Read(unit,*) 
	   Read(unit,*)
	end do
	NE = NE

	Read(unit,*)         ! Nodes		 
	Read(unit,*)
	Read(unit,*)
	Read(unit,*)
	Read(unit,*) NN
	Allocate( npos(3,NN) )
    do i=1,NN
	   Read(unit,*) (npos(j,i), j=1,3) 
	end do
    
    Close(unit)

Return
End subroutine VOL_read_FE

!===============================================================
!  get:   ND,NN,npos, NC,C2N,BC_Cell
! 
Subroutine NEU_read_FE (filename)
Implicit None
Character (Len=*) :: filename

Integer :: unit,ios, i,j

    unit = 10
    Open (unit, File=filename//".neu", Iostat=ios, Form="Formatted", Action="Read")
	if (ios /= 0) then
		Print *, "Error opening .neu file. Error code = ", ios, filename//".neu" 
		stop 
	end if

	Read(unit,*) NN			! Nodes
	Allocate( npos(3,NN) )
    do i=1,NN
	   Read(unit,*) (npos(j,i), j=1,3) 
	end do


	Read(unit,*) NC			! cell elements
	if (NC > 0) then
      ND = 3
	  NCN = ND+1
	  Allocate( C2N(NCN,NC), BC_Cell(NC) )
      do i=1,NC
	     Read(unit,*) BC_Cell(i), (C2N(j,i), j=1,NCN) 
	  end do
	end if

    if (NC == 0) then       ! B faces  (2D elements)
      ND = 2
	  NCN = ND+1
	  Read(unit,*) NC			
	  Allocate( C2N(NCN,NC), BC_Cell(NC) )
      do i=1,NC
	     Read(unit,*) BC_Cell(i), (C2N(j,i), j=1,NCN) 
	  end do
	end if

    Close(unit)


Return
End subroutine NEU_read_FE

!===============================================================
!  write FV
!===============================================================
! Put	NFN,NF,F2C,F2N, NFBC,fBC_pt, BC_Face
! 
Subroutine Write_FV_Con (unit)
Implicit None
Integer :: unit

Integer :: i,j

	Write(unit,*) "face connectivity [F2C(1:2), F2N(1:NFN)]"
	Write(unit,*) 2+NFN," 1  ", NF			! Face Connectivity
    do i=1,NF
	   Write(unit,*) (F2C(j,i), j=1,2), (F2N(j,i), j=1,NFN)
	end do

	Write(unit,*) "face BC pointers and info [pt(1:2), info]"
	Write(unit,*) "3   1  ", NFBC				! Face BCs
    do i=1,NFBC
	   Write(unit,*) (fBC_pt(j,i), j=1,3)
	end do		
					
	Write(unit,*) "face BC data  [integer data item]"
	Write(unit,*) "1   1  ", fBC_pt(2,NFBC)	! BCs data
    do i=1,fBC_pt(2,NFBC)
	   Write(unit,*) BC_Face(i)
	end do

	Write(unit,*) " "

Return
End subroutine Write_FV_Con

!===============================================================
! Put	NN,npos, NNBC,nBC_pt, BC_Node
!
Subroutine Write_FV_Pos (unit)
Implicit None
Integer :: unit

Integer :: i,j

	Write(unit,*) "node positions [x,y,{z}]"			
	Write(unit,*) "3   1  ", NN			! Nodes
    do i=1,NN
	   Write(unit,*) (npos(j,i), j=1,3) 
	end do

	Write(unit,*) "node BC pointers and info [pt(1:2), info]"
	Write(unit,*) "3   1  ", NNBC				! Face BCs
    do i=1,NNBC
	   Write(unit,*) (nBC_pt(j,i), j=1,3)
	end do		
					
	Write(unit,*) "node BC data  [integer data item]"
	Write(unit,*) "1   1  ", nBC_pt(2,NNBC)	! BCs data
    do i=1,nBC_pt(2,NNBC)
	   Write(unit,*) BC_Node(i)
	end do

	Write(unit,*) " "

Return
End subroutine Write_FV_Pos

!===============================================================
! Put NCN,NC, BC_Cell,C2N
!
Subroutine Write_FV_Mat (unit)
Implicit None
Integer :: unit

Integer :: i,j

	Write(unit,*) "cell info   [material and C2N(NCN)]"		
	Write(unit,*) 1+NCN," 1 ", NC				
    do i=1,NC
	   Write(unit,*) BC_Cell(i),(C2N(j,i), j=1,NCN)
	end do

	Write(unit,*) " "

Return
End subroutine Write_FV_Mat

!===============================================================
!  read FV
!===============================================================
! Get	NFN,NF,F2C,F2N, NFBC,fBC_pt, BC_Face
! 
Subroutine Read_FV_Con (unit)
Implicit None
Integer :: unit

Integer :: i,j, i1,i2, j2

	Read(unit,*) ! "face connectivity [F2C(1:2), F2N(1:NFN)]"
	Read(unit,*) NFN, i1, NF			! Face Connectivity
	NFN = NFN-2
	Allocate ( F2C(2,i1:NF), F2N(NFN,i1:NF) )
    do i=i1,NF
	   Read(unit,*) (F2C(j,i), j=1,2), (F2N(j,i), j=1,NFN)
	end do

	Read(unit,*) ! "face BC pointers and info [pt(1:2), info]"
	Read(unit,*) j2, i1 , NFBC				! Face BCs
	Allocate ( fBC_pt(j2,i1:NFBC) )
    do i=i1,NFBC
	   Read(unit,*) (fBC_pt(j,i), j=1,j2)
	end do		
					
	Read(unit,*) !"face BC data  [integer data item]"
	Read(unit,*) j2, i1 , i2	! BCs data
	Allocate ( BC_Face(i1:i2) )
    do i=i1,i2
	   Read(unit,*) BC_Face(i)
	end do

	Read(unit,*) !" "

Return
End subroutine Read_FV_Con
!
!===============================================================
!
Subroutine Deallocate_FV_Con ()
Implicit None

	Deallocate ( F2C, F2N )
	Deallocate (fBC_pt, BC_Face )

Return
End Subroutine Deallocate_FV_Con

!===============================================================
! Get	NN,npos, NNBC,nBC_pt, BC_Node
!
Subroutine Read_FV_Pos (unit)
Implicit None
Integer :: unit

Integer :: i,j, i1,i2, j2

	Read(unit,*) ! "node positions [x,y,{z}]"			
	Read(unit,*) j2, i1, NN			! Nodes
	Allocate ( npos(j2,i1:NN) )
    do i=i1,NN
	   Read(unit,*) (npos(j,i), j=1,j2) 
	end do

	Read(unit,*) !"node BC pointers and info [pt(1:2), info]"
	Read(unit,*) j2, i1, NNBC				! Face BCs
	Allocate ( nBC_pt(j2,i1:NNBC) )
    do i=i1,NNBC
	   Read(unit,*) (nBC_pt(j,i), j=1,j2)
	end do		
					
	Read(unit,*) !"node BC data  [integer data item]"
	Read(unit,*) j2, i1, i2	! BCs data
	Allocate ( BC_Node(i1:i2) )
    do i=i1,i2
	   Read(unit,*) BC_Node(i)
	end do

	Read(unit,*) !" "

Return
End Subroutine Read_FV_Pos
!
!===============================================================
!
Subroutine Deallocate_FV_Pos ()
Implicit None

	Deallocate ( npos )
	Deallocate ( nBC_pt, BC_Node )

Return
End Subroutine Deallocate_FV_Pos

!===============================================================
! Get NCN,NC, BC_Cell,C2N
!
Subroutine Read_FV_Mat (unit)
Implicit None
Integer :: unit

Integer :: i,j, i1

	Read(unit,*) !"cell info   [material and C2N(NCN)]"		
	Read(unit,*) NCN, i1, NC
	NCN = NCN-1
	Allocate ( BC_Cell(i1:NC), C2N(NCN,i1:NC) )				
    do i=i1,NC
	   Read(unit,*) BC_Cell(i),(C2N(j,i), j=1,NCN)
	end do

	Read(unit,*) !" "

Return
End subroutine Read_FV_Mat
!
!===============================================================
!
Subroutine Deallocate_FV_Mat ()
Implicit None

	Deallocate ( C2N )
	Deallocate ( BC_Cell )

Return
End Subroutine Deallocate_FV_Mat

!===============================================================
!  get new connectivity format
!===============================================================
! assumes simplices  (maybe can generalize)
!  C2N => {NFN},NF,F2C,F2N
!
Subroutine Convert ()
Implicit None

Integer :: F2Ctmp(2,NC*(ND+2)/2), F2Ntmp(ND,NC*(ND+2)/2), F2Nt(ND+1,(ND+1)*NC)
Integer :: match((ND+1)*NC)
Integer :: i,j,k, m,fnum, FT 



! lots of faces  (some duplicates)
    fnum = 0
    do i=1,NC  !for each cell
	  do j=1,ND+1   !for each face of the cell
	    fnum = fnum+1
		do k=1,ND   !get three nodes in order
		  m = Mod(k+j-2,ND+1) + 1
	      F2Nt(k,fnum) = C2N(m,i)
		end do
	    F2Nt(ND+1,fnum) = i      !cell number
      end do 	  
    end do
    FT = fnum

! search for duplicates
    F2C = 0             !zero everything (not just 2)
	fnum = 0
    do i=1,FT

      do j=1,fnum
	    match(i) = 0
	    do k=1,ND
		do m=1,ND
          if ( F2Nt(m,i) == F2Ntmp(k,j) ) match(i) = match(i)+1
        end do
		end do

	    if (match(i) == ND) then
           F2Ctmp(2,j) = F2Nt(ND+1,i)
		   exit                 !exit
		end if
	  end do

	  if (match(i) < ND) then
	    fnum = fnum+1
		do j=1,ND
		  F2Ntmp(j,fnum) = F2Nt(j,i)
        end do
	    F2Ctmp(1,fnum) = F2Nt(ND+1,i)
	  end if

    end do
	NF = fnum

    Allocate ( F2C(2,NF), F2N(ND,NF) )
	NFN = ND  !Simplicies
	F2C = F2Ctmp(:,1:NF)
	F2N = F2Ntmp(:,1:NF)

Return
End subroutine Convert

!===============================================================
! Not debugged
!  F2C,F2N => NC,C2N
!
Subroutine Revert()
Implicit None

Logical, Allocatable, DImension(:) :: match     !NC is unknown
Integer, Allocatable, DImension(:) :: Ccnt
Integer, Allocatable, DImension(:,:) :: C2Nt
Integer :: i,j,k, c1,c2

! lots of faces  (some duplicates)

    NC = MaxVal(F2C)
	Allocate ( Ccnt(NC), C2Nt(ND*(ND+1),NC), match(NC), C2N(ND+1,NC) )

    Ccnt = 0
	do i=1,NF
      c1 = F2C(1,i)
      c2 = F2C(2,i)
	  	  
      do j=1,ND
        C2Nt(Ccnt(c1)+j,c1) = F2N(j,i)
        C2Nt(Ccnt(c2)+j,c2) = F2N(j,i)
	  end do
	  Ccnt(c1) = Ccnt(c1)+ND
	  Ccnt(c2) = Ccnt(c2)+ND
    end do

    Ccnt = 0
    do i=1,NC

      do j=1,ND*(ND+1)

        match(i) = .false.
	    do k=1,Ccnt(i)
          if ( C2Nt(j,i) == C2N(k,i) ) then
		    match(i) = .true.
		  end if
		end do

		if (.not. match(i)) then
		  Ccnt(i) = Ccnt(i) + 1
          C2N(Ccnt(i),i) = C2Nt(j,i)
		end if

		if (Ccnt(i) == ND+1) exit 
      end do

    end do

	Deallocate ( Ccnt, C2Nt, match )
Return
End subroutine Revert

!===============================================================
!===============================================================


End Module FE_mesh_IO